import pytest
from unittest.mock import patch
from pynonymizer.database.exceptions import UnknownDatabaseTypeError
from pynonymizer.database import get_provider

@patch("pynonymizer.database.MySqlProvider")
def test_get_mysql_provider(mysql_provider):
    """
    given "mysql" type, get_provider should get a MySqlProvider
    """
    provider = get_provider("mysql", "db_host", "db_user", "db_password", "db_name")

    assert provider == mysql_provider.return_value


def test_get_unknown_provider():
    """
    get_provider should raise UnknownDatabaseTypeError if given unknown type
    """
    with pytest.raises(UnknownDatabaseTypeError) as e_info:
        provider = get_provider("UNKNOWN TYPE", "db_host", "db_user", "db_password", "db_name")


from unittest.mock import patch
import pytest
from pynonymizer.log import get_logger, get_default_logger


@patch("logging.getLogger")
def test_get_default_logger(get_logger_mock):
    logger = get_default_logger()
    get_logger_mock.assert_called_once_with("pynonymizer")


@patch("logging.getLogger")
def test_simple_name(get_logger_mock):
    logger = get_logger("test")
    get_logger_mock.assert_called_once_with("pynonymizer.test")


@patch("logging.getLogger")
def test_multipart_name(get_logger_mock):
    logger = get_logger(["a", "b", "cd", "ef", "ghi"])
    get_logger_mock.assert_called_once_with("pynonymizer.a.b.cd.ef.ghi")


@patch("logging.getLogger")
def test_no_logger_name_array(get_logger_mock):
    with pytest.raises(ValueError):
        logger = get_logger([])


@patch("logging.getLogger")
def test_no_logger_name_str(get_logger_mock):
    with pytest.raises(ValueError):
        logger = get_logger("")


from pynonymizer.database import get_temp_db_name

def test_uniqueness():
    """
    Generate 100 temp db names, and test if they're unique.
    """
    db_names = [get_temp_db_name("filename.sql") for i in range(100)]
    assert len(db_names) == len(set(db_names))




import pytest
from pynonymizer.database.exceptions import DatabaseProviderError, DependencyError, UnsupportedTableStrategyError, UnsupportedColumnStrategyError


def test_database_provider_error():
    error = DatabaseProviderError()


def test_dependency_error():
    error = DependencyError("mysqldump", "error message")
    assert error.name == "mysqldump"


def test_unsupported_table_strategy():
    error = UnsupportedTableStrategyError("error message")


def test_unsupport_column_strategy_error():
    error = UnsupportedColumnStrategyError("COLUMN_STRATEGY_TYPE")



import unittest
from unittest.mock import Mock
import pytest
from pynonymizer.fake import FakeDataType
from pynonymizer.database.exceptions import UnsupportedColumnStrategyError
from pynonymizer.strategy.table import UpdateColumnsTableStrategy
from pynonymizer.strategy.update_column import (
    FakeUpdateColumnStrategy,
    EmptyUpdateColumnStrategy,
    UniqueLoginUpdateColumnStrategy,
    UniqueEmailUpdateColumnStrategy,
    LiteralUpdateColumnStrategy
)
import pynonymizer.database.postgres.query_factory as query_factory

"""
These tests are brittle and based on the actual SQL generatedColumnStrategyTypes. 
The sentiment is to test the 'meaning' of the SQL, rather than the actual formatting, so it may be prudent to replace
these tests with some form of parsing or pattern matching. 

The general idea, however, is that by keeping the queryfactory separate from the provider, it will not change often,
and the sql returned should be very stable.
"""


def test_get_truncate_table(simple_strategy_trunc):
    assert query_factory.get_truncate_table(simple_strategy_trunc) == "TRUNCATE TABLE truncate_table CASCADE;"


def test_get_schema_truncate_table(simple_strategy_schema_trunc):
    assert query_factory.get_truncate_table(simple_strategy_schema_trunc) == "TRUNCATE TABLE schema.truncate_schema_table CASCADE;"


def test_get_drop_seed_table():
    assert query_factory.get_drop_seed_table("seed_table") == "DROP TABLE IF EXISTS seed_table;"


def test_get_create_database():
    assert query_factory.get_create_database("test_database") == "CREATE DATABASE test_database;"


def test_get_drop_database():
    assert query_factory.get_drop_database("test_database") == [
        "SELECT pid, pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname = 'test_database' AND pid != pg_backend_pid();",
        "DROP DATABASE IF EXISTS test_database;"
    ]


def test_get_dumpsize_estimate():
    assert query_factory.get_dumpsize_estimate("test") == "SELECT 1;"

@pytest.fixture
def str_fake_column_generator():
    return Mock(
            get_data_type=Mock(return_value=FakeDataType.STRING),
            get_value=Mock(return_value="test_value")
        )
@pytest.fixture
def int_fake_column_generator():
    return Mock(
        get_data_type=Mock(return_value=FakeDataType.INT),
        get_value=Mock(return_value=645)
    )


@pytest.fixture
def fake_update_column_str_first_name(str_fake_column_generator):
    return FakeUpdateColumnStrategy("test_column1",str_fake_column_generator, "first_name")


@pytest.fixture
def fake_update_column_int_last_name(int_fake_column_generator):
    return FakeUpdateColumnStrategy("test_column2",int_fake_column_generator, "last_name")


@pytest.fixture
def empty_strategy():
    return EmptyUpdateColumnStrategy("test_column3")


@pytest.fixture
def ulogin_strategy():
    return UniqueLoginUpdateColumnStrategy("test_column4")


@pytest.fixture
def uemail_strategy():
    return UniqueEmailUpdateColumnStrategy("test_column5")


@pytest.fixture
def literal_strategy():
    return LiteralUpdateColumnStrategy("test_column6", value="RANDOM()")


@pytest.fixture
def database_strategy():
    pass


@pytest.fixture
def qualifier_column_map(fake_update_column_str_first_name,fake_update_column_int_last_name,empty_strategy,ulogin_strategy, uemail_strategy,literal_strategy):
    return {
        "first_name": fake_update_column_str_first_name,
        "last_name": fake_update_column_int_last_name,
        "first_name_test_arg_5": fake_update_column_str_first_name
    }


@pytest.fixture
def column_strategy_list(fake_update_column_str_first_name,fake_update_column_int_last_name,empty_strategy,ulogin_strategy, uemail_strategy,literal_strategy):
    return [
        fake_update_column_str_first_name,
        fake_update_column_int_last_name,
        empty_strategy,
        ulogin_strategy,
        uemail_strategy,
        literal_strategy
    ]


@pytest.fixture
def update_table_strategy(column_strategy_list):
    return UpdateColumnsTableStrategy("table_name", column_strategy_list)


@pytest.fixture
def update_table_strategy_unknown(unsupported_column_strategy):
    return UpdateColumnsTableStrategy("invalid_table", [unsupported_column_strategy])


@pytest.fixture
def unsupported_column_strategy():
    column = UniqueLoginUpdateColumnStrategy("column_name")
    column.strategy_type = "NOT_SUPPORTED"

    return column


def test_get_insert_seed_row(qualifier_column_map):
    insert_seed_row = query_factory.get_insert_seed_row("seed_table", qualifier_column_map)

    assert insert_seed_row == "INSERT INTO seed_table(first_name,last_name,first_name_test_arg_5) " \
                              "VALUES ('test_value',645,'test_value');"


def test_get_create_seed_table(qualifier_column_map):
    assert query_factory.get_create_seed_table("seed_table", qualifier_column_map) == "CREATE TABLE seed_table (first_name VARCHAR(65535),last_name INT,first_name_test_arg_5 VARCHAR(65535));"


def test_get_create_seed_table_no_columns():
    """
    get_create_seed_table should error when presented with no columns
    """
    with pytest.raises(ValueError) as e_info:
        query_factory.get_create_seed_table("seed_table", {})


def test_get_update_table_unsupported_column_type(update_table_strategy_unknown):
    """
    get_update_table should raise UnsupportedColumnStrategyError if presented with an unsupported column type
    """
    with pytest.raises(UnsupportedColumnStrategyError):
        query_factory.get_update_table("seed_table", update_table_strategy_unknown)


def test_get_update_table_fake_column(column_strategy_list):
    update_table_all = query_factory.get_update_table("seed_table", UpdateColumnsTableStrategy("anon_table", column_strategy_list))

    assert update_table_all == [
            "UPDATE anon_table AS \"updatetarget\" SET "
            "test_column1 = ( SELECT first_name FROM seed_table WHERE \"updatetarget\"=\"updatetarget\" ORDER BY RANDOM() LIMIT 1),"
            "test_column2 = ( SELECT last_name FROM seed_table WHERE \"updatetarget\"=\"updatetarget\" ORDER BY RANDOM() LIMIT 1),"
            "test_column3 = (''),"
            "test_column4 = ( SELECT md5(random()::text) WHERE \"updatetarget\"=\"updatetarget\" ),"
            "test_column5 = ( SELECT CONCAT(md5(random()::text), '@', md5(random()::text), '.com') WHERE \"updatetarget\"=\"updatetarget\" ),"
            "test_column6 = RANDOM();"
            ]


def test_get_update_table_literal(literal_strategy):

    result_queries = query_factory.get_update_table("seed_table", UpdateColumnsTableStrategy("anon_table", [
        LiteralUpdateColumnStrategy("literal_column", "RANDOM()")
    ]))

    assert result_queries == [
        "UPDATE anon_table AS \"updatetarget\" SET literal_column = RANDOM();"
    ]


# Minimum required module file
# ----------------------------
# Because pynonymizer.version is imported by setup.py, this needs to stay empty, lest it cause import errors pre-install
# However, this makes sense because this is the base package for a tool, and we have other entry points anyway.
#
# So if you're thinking about adding application code here, you're probably in the wrong place.



import pytest
from unittest.mock import Mock, patch, mock_open
from pynonymizer.strategy.config import read_config, UnknownConfigTypeError

test_path_examples = [
    "",
    "complex/multi/part/path/test/",
    "complex\\multi\\part\\path\\test\\",
    "/absolute/path/test/",
    "C:\\absolute\\path\\test\\"
]


@pytest.mark.parametrize("path_example", test_path_examples)
@patch("builtins.open")
@patch("pynonymizer.strategy.config.load_yaml")
@patch("pynonymizer.strategy.config.load_json")
def test_read_config_json(load_json, load_yaml, open, path_example):
    filepath = path_example + "test.json"
    config = read_config(filepath)
    load_json.assert_called_once_with( open().__enter__().read() )
    assert config == load_json()


@pytest.mark.parametrize("path_example", test_path_examples)
@patch("builtins.open")
@patch("pynonymizer.strategy.config.load_yaml")
@patch("pynonymizer.strategy.config.load_json")
def test_read_config_yml(load_json, load_yaml, open, path_example):
    filepath = path_example + "test.yml"
    config = read_config(filepath)
    load_yaml.assert_called_once_with(open().__enter__())
    assert config == load_yaml()


@pytest.mark.parametrize("path_example", test_path_examples)
@patch("builtins.open")
@patch("pynonymizer.strategy.config.load_yaml")
@patch("pynonymizer.strategy.config.load_json")
def test_read_config_yaml(load_json, load_yaml, open, path_example):
    filepath = path_example + "test.yaml"
    config = read_config(filepath)
    load_yaml.assert_called_once_with(open().__enter__())
    assert config == load_yaml()


@pytest.mark.parametrize("path_example", test_path_examples)
@patch("builtins.open")
@patch("pynonymizer.strategy.config.load_yaml")
@patch("pynonymizer.strategy.config.load_json")
def test_read_config_unknown(load_json, load_yaml, open, path_example):
    with pytest.raises(UnknownConfigTypeError):
        filepath = path_example + "test.bbs"
        config = read_config(filepath)


import unittest
from unittest.mock import Mock, patch
import pytest

from pynonymizer.fake import FakeColumnGenerator, FakeDataType, UnsupportedFakeTypeError
from faker import Faker


class FakeColumnGeneratorTests(unittest.TestCase):
    def setUp(self):
        self.generator = FakeColumnGenerator()

    def test_supports_supported(self):
        assert self.generator.supports("first_name") is True

    def test_supports_unsupported_method(self):
        assert self.generator.supports("NOT A VALID METHOD") is False

    def test_supports_unsupported_args(self):
        assert self.generator.supports("first_name", {"arg1": 2, "arg2": 3}) is False

    def test_get_data_type_known_column(self):
        assert self.generator.get_data_type("date") == FakeDataType.DATE

    def test_get_data_type_unknown_column(self):
        assert self.generator.get_data_type("user_agent") == FakeDataType.STRING

    @patch("pynonymizer.fake.Faker")
    def test_get_value_unsupported(self, faker):
        class FakeMock:
            def ean(unmatched_args):
                pass

        faker.return_value = FakeMock()
        faked_generator = FakeColumnGenerator()
        with pytest.raises(UnsupportedFakeTypeError):
            faked_generator.get_value("NOT A VALID METHOD")

    @patch("pynonymizer.fake.Faker")
    def test_get_value_unsupported_args(self, faker):
        class FakeMock:
            def ean(unmatched_args):
                pass

        faker.return_value = FakeMock()
        faked_generator = FakeColumnGenerator()
        with pytest.raises(UnsupportedFakeTypeError):
            faked_generator.get_value("ean", {"not_valid_length": 13})



import logging
import os

def get_default_logger():
    """
    Get the default logging instance
    :return:
    """
    return logging.getLogger("pynonymizer")


def get_logger(name):
    """
    Get a child logger of the default logger
    :param name: expects an arbitrary name or list of names to form into heirarchy
    :return:
    """

    if isinstance(name, str):
        child_logger_name = name
    else:
        child_logger_name = ".".join(name)

    if len(child_logger_name) < 1:
        raise ValueError("No logger name given.")

    return logging.getLogger(f"pynonymizer.{child_logger_name}")


def init_logging():
    # init default logger
    default_logger = get_default_logger()

    default_logger.setLevel(logging.DEBUG)
    default_file_formatter = logging.Formatter('[%(asctime)-15s|%(levelname)s|%(name)s] %(message)s')
    default_console_formatter = logging.Formatter('%(message)s')

    # Add default warn handler to console
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(default_console_formatter)
    default_logger.addHandler(console_handler)

    # Add default file handler when running in production
    if os.getenv("env_type") == "production":
        file_handler = logging.FileHandler('pynonymizer.log')
        file_handler.setLevel(logging.INFO)
        file_handler.setFormatter(default_file_formatter)
        default_logger.addHandler(file_handler)


init_logging()
